const EXPORTED_SYMBOLS = ["SetWindow", "getAboutUriChannel",
  "utilEx", // about:multifox
  "Profile" // about:multifox
]
const Cc = Components.classes;
const Ci = Components.interfaces;

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
Components.utils.import("resource://multifox-modules/util.js");

util0.log("===>LOADING main.js");

const SetWindow = {
  applyRestore: function(win) {
    // restore: window is first configured by SetWindow.inheritId (via nsIWindowMediator.addListener)
    util0.log('applyRestore ');
    Profile.saveRestore(win);
  },

  newId: function(win) {
    var id;
    if (shouldBeDefault(win)) {
      id = DefaultIdentity;
    } else {
      id = Profile.lowerAvailableId(win);
    }
    util0.log('newIdentity '+id);
    Profile.save(win, id);
  },

  inheritId: function(newWin, prevWin) {
    var id;
    if (shouldBeDefault(newWin)) {
      id = DefaultIdentity;
    } else {
      if (prevWin) {
        id = Profile.getIdentity(prevWin.document);
      } else {
        id = UnknownIdentity;
      }
    }
    util0.log('inheritedIdentity '+id);
    Profile.save(newWin, id);
  }
};


function shouldBeDefault(win) {
  // popup opened by an extension (like GTB)
  var crHidden = win.document.documentElement.getAttribute("chromehidden");
  return crHidden.indexOf("location") > -1;
}


function onUnloadChromeWindow(evt) {
  var win = evt.currentTarget;
  unregisterWindow(win);
}


function registerWindow(win, profileId) {
  var doc = win.document;

  if (Profile.getIdentity(doc) !== DefaultIdentity) {
    utilEx.thrw.myThrow('already registered win ' + Profile.getIdentity(doc));
  }

  createIcon(doc, profileId);

  util0.log("registerWindow " + profileId);
  if (profileId === DefaultIdentity) {
    return;
  }

  if (util0.setHttpObservers.isActive === false) {
    // first multifox window!
    init();
  }

  var cook = Cookies.regWindow(profileId);
  XblContent.registerWindow2(win);

  win.addEventListener("MultifoxContentEvent_CookieGetter", cook.cookieGetter, false, true);
  win.addEventListener("MultifoxContentEvent_CookieSetter", cook.cookieSetter, false, true);
  win.addEventListener("MultifoxContentEvent_FeatureNotSupported", onNotSupported, false, true);

  // {xxxWTF} some fromXbl* listeners are not called when
  // there are "unload" listeners with useCapture=true. o_O
  // But they are called if listener is an anonymous function.
  // it doesn't work properly for some reason that's beyond me
  win.addEventListener("unload", onUnloadChromeWindow, false);
}


// should keep id for session restore?
function unregisterWindow(win) {
  var doc = win.document;

  var idw = Profile.getIdentity(doc);
  util0.log('unreg as ' + idw);

  if (idw === DefaultIdentity) {
    // nothing to unregister
    return;
  }


  XblContent.unregisterWindow2(win);
  var sessions = Profile.activeIdentities(win);
  var cook = Cookies.removeWindow(idw, sessions);

  win.removeEventListener("MultifoxContentEvent_CookieGetter", cook.cookieGetter, false);
  win.removeEventListener("MultifoxContentEvent_CookieSetter", cook.cookieSetter, false);
  win.removeEventListener("MultifoxContentEvent_FeatureNotSupported", onNotSupported, false);

  var onlyDefault = (sessions.length === 1) && (sessions[0] === DefaultIdentity);
  if (onlyDefault) {
    // no more multifox windows
    shutdown();
  }
  win.removeEventListener("unload", onUnloadChromeWindow, false);
}


function init() {
  util0.log('init');
  Cookies.init3();
  util0.setHttpObservers.enable(httpObservers.request, httpObservers.response);
  XblContent.init2();
  toggleDefaultWindowUI(true);
}

function shutdown() {
  util0.log('shutdown');
  util0.setHttpObservers.disable();
  XblContent.shutdown2();
  Cookies.shutdown3();
  toggleDefaultWindowUI(false);
}

function onNotSupported(evt) {
  evt.stopPropagation();
  if (evt.data === "__defineSetter__") {
    var doc = evt.target;
    utilEx.dbg('__defineSetter__?',
               doc.location,
               'title[' + doc.title + ']',
               doc.__defineSetter__,
               doc.wrappedJSObject.__defineSetter__,
               utilEx.findBrowser(doc.defaultView),
               utilEx.findBrowser2(doc.defaultView)
    );
    return;
  }
  Components.utils.import("resource://multifox-modules/error.js");
  showError(evt.target.defaultView, evt.data);
}
const UnknownIdentity = 0;
const DefaultIdentity = 1;

const Profile = {

  getIdentity: function(doc) {
    var icon = getIconNode(doc);
    if (icon === null) {
      var type = doc.documentElement.getAttribute("windowtype");
      if (type === "navigator:view-source") {
        var winOpener = doc.defaultView.opener;
        if (winOpener) {
          icon = getIconNode(winOpener.document);
        }
        util0.log("view-source opener=" + winOpener + (icon ? icon.getUserData("multifox-dom-identity-id") : "[?]"));
      }
    }

    if (icon === null) {
      return DefaultIdentity;
    } else {
      var id = icon.getUserData("multifox-dom-identity-id");
      return this.toInt(id);
    }
  },

  save: function(win, id) {
    util0.log("save "+id);
    if (Profile.getIdentity(win.document) !== DefaultIdentity) {
      unregisterWindow(win);
      removeIcon(win.document);
    }

    registerWindow(win, id);
    new SaveToSessionStore(win.document);
  },

  saveRestore: function(win) { //saveToIcon
    win.document.setUserData("multifox-dom-restore-fired", "ok", null);

    if (Profile.getIdentity(win.document) !== DefaultIdentity) {
      //util0.log("saveRestore unreg ");
      unregisterWindow(win);
      removeIcon(win.document);
    } else {
      // DefaultIdentity
      removeIcon(win.document);
    }

    var stringId = Cc["@mozilla.org/browser/sessionstore;1"]
                    .getService(Ci.nsISessionStore)
                    .getWindowValue(win, "multifox-dom-identity-id");
    var id = Profile.toInt(stringId);
    if (id < DefaultIdentity) { //UnknownIdentity?
      id = DefaultIdentity;
    }
    registerWindow(win, id);
  },

  activeIdentities: function(ignoreWin) {
    var winEnum = utilEx.browserWindowsEnum();
    var arr = [];
    while (winEnum.hasMoreElements()) {
      var win = winEnum.getNext();
      if (ignoreWin !== win) {
        var id = Profile.getIdentity(win.document);
        //util0.log('activeIdentities loop='+id);
        if (arr.indexOf(id) === -1) {
          arr.push(id);
        }
      }
    }
    return arr;
  },

  lowerAvailableId: function(ignoreWin) {
    var arr = this.activeIdentities(ignoreWin); //ignore win -- it doesn't have a session id yet
    var id = DefaultIdentity;
    while (arr.indexOf(id) > -1) {
      id++;
    }
    return id; // id is available
  },

  toInt: function(str) {
    var rv = parseInt(str, 10);
    return isNaN(rv) ? DefaultIdentity : rv;
  },

  toString: function(id) {
    switch (id) {
      case UnknownIdentity:
        return "\u221e"; // ∞
      default:
        return id.toString();
    }
  }

};


function SaveToSessionStore(doc) {
  this._doc = doc;
  this._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
  this._timer.init(this, 1300, Ci.nsITimer.TYPE_ONE_SHOT);
}

SaveToSessionStore.prototype = {
  _doc: null,
  _timer: null,

  _restored: function() {
    return this._doc.getUserData("multifox-dom-restore-fired") === "ok";
  },

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference]),

  observe: function(aSubject, aTopic, aData) {//nsITimer "timer-callback" null
    util0.log('SaveToSessionStore timer!');
    var doc = this._doc;
    if (doc.defaultView === null) {
      util0.log("_syncToSessionStore window closed");
      return;
    }

    if (this._restored()) {
      // cancel, icon was updated by saveRestore
      util0.log("_syncToSessionStore CANCELED icon=" +
                Profile.getIdentity(doc) +
                ' sstore=' +
                Cc["@mozilla.org/browser/sessionstore;1"].getService(Ci.nsISessionStore).getWindowValue(doc.defaultView, "multifox-dom-identity-id"));
      return;
    }

    // ok, restore nao ocorreu (pode nem ocorrer)
    var ss = Cc["@mozilla.org/browser/sessionstore;1"].getService(Ci.nsISessionStore);
    var val = Profile.getIdentity(doc);

    try {
      // overwrite any previous value if called twice
      ss.setWindowValue(doc.defaultView, "multifox-dom-identity-id", val);
    } catch (ex) {
      // keep trying
      util0.log("_syncToSessionStore FAIL "+val+ doc.defaultView.state + '\n'+ex);
      this._timer.init(this, 300, Ci.nsITimer.TYPE_ONE_SHOT);
      return;
    }

    if (val <= DefaultIdentity) { // UnknownIdentity OR DefaultIdentity
      ss.deleteWindowValue(doc.defaultView, "multifox-dom-identity-id");
    }

    //this._timer.init(this, 40000, Ci.nsITimer.TYPE_ONE_SHOT);
    util0.log("_syncToSessionStore OK=" + Profile.getIdentity(doc));
  }
}

// show/hide [1] "icon"
function toggleDefaultWindowUI(show) {
  var winEnum = utilEx.browserWindowsEnum();
  while (winEnum.hasMoreElements()) {
    var doc = winEnum.getNext().document;
    if (Profile.getIdentity(doc) === DefaultIdentity) {
      if (show) {
        createIcon(doc, DefaultIdentity);
      } else {
        removeIcon(doc);
      }
    }
  }
}

function getIconNode(doc) {
  if (!doc) {
    utilEx.thrw.myThrow("getIconNode "+doc);
  }
  return doc.getElementById("multifox-icon");
}


function removeIcon(doc) {
  util0.log('removeIcon');
  var icon = getIconNode(doc);
  if (icon) {
    icon.parentNode.removeChild(icon);
  }
}


/*
<box>
  <label>

*/
function createIcon(doc, profileId) {
  if (getIconNode(doc)) {
    utilEx.thrw.myThrow('createIcon ERR new=' + profileId + " current=" + Profile.getIdentity(doc));
  }
  /*
  var ub = doc.getElementById("urlbar");//textbox
  var dropMarker = doc.getAnonymousElementByAttribute(ub, "anonid", "popupset");
  var pn = doc.getBindingParent(dropMarker);
  util0.log('pn=' + pn.tagName);
  try {
    var icon2 = dropMarker.insertBefore(doc.createElement("box"), dropMarker.firstChild);
    //var spc = pn.appendChild(doc.createElement("spacer"));
    //spc.setAttribute('flex', '1');
    var icon2 = pn.appendChild(doc.createElement("box"));
    var lb2 = icon2.appendChild(doc.createElement("label"));
    lb2.setAttribute("value", profileId);
    var s = lb2.style;
    s.color = "red";
  } catch (ex) {
    utilEx.thrw.myThrow('ico ' + ex);
  }
  */

  util0.log('createIcon ' + profileId);
  var bar = doc.getElementById("urlbar-icons");
  var icon = bar.appendChild(doc.createElement("box"));
  var lb = icon.appendChild(doc.createElement("label"));

  icon.setAttribute("id", "multifox-icon");
  icon.setAttribute("tooltiptext", "Multifox (BETA)");
  icon.style.MozBorderRadius = "2px";
  icon.style.paddingLeft = "6px";
  icon.style.paddingRight = "6px";
  icon.style.background = "green url(chrome://multifox/content/icon.png) repeat-x";

  lb.setAttribute("value", Profile.toString(profileId));
  var s = lb.style;
  s.color = "white";
  s.fontWeight = "bold";
  s.fontStyle = "normal";
  s.textRendering = "optimizelegibility";
  s.textShadow = "1px 1px 1px black";

  initIconNormal(icon);
  icon.setUserData("multifox-dom-identity-id", profileId === DefaultIdentity ? null : profileId, null);
}

function openMultifoxPopup(evt) {
  var icon = evt.currentTarget;//or this
  var doc = icon.ownerDocument;
  var win = doc.defaultView;

  var profileId = Profile.getIdentity(doc);
  var p1 = util0.getText("icon.panel.p1.label", Profile.toString(profileId));
  var p2 = util0.getText("icon.panel.p2.label", utilEx.appName(), "Multifox (BETA)");
  var p3 = util0.getText("icon.panel.p3.label");


  var panel = doc.getElementById("multifox-popup");
  if (panel) {
    //bug
    win.alert('OPENMULTIFOXPOPUP ==>HIDING '+panel.state);
    panel.hidePopup();
    return;
  }

  panel = doc.getElementById("mainPopupSet").appendChild(doc.createElement("panel"));
  var hx = panel.appendChild(doc.createElement("hbox"));


  var img = hx.appendChild(doc.createElement("image"));
  var desc1 = panel.appendChild(doc.createElement("description"));
  var desc3 = panel.appendChild(doc.createElement("description"));
  var desc2 = panel.appendChild(doc.createElement("description"));

  img.setAttribute("src", "chrome://multifox/content/logo-popup.png");
  desc1.appendChild(doc.createTextNode(p1));
  desc2.appendChild(doc.createTextNode(p2));
  desc3.appendChild(doc.createTextNode(p3));
  desc1.style.fontWeight = "bold";


  if (profileId === UnknownIdentity) {
    desc2.hidden = true;
    desc3.hidden = true;
  }

  if (profileId !== DefaultIdentity) {
    desc2.hidden = true;
  }


  // about button
  var h = panel.appendChild(doc.createElement("hbox"));
  var spc = h.appendChild(doc.createElement("spacer"));
  spc.flex = 1;
  var but = h.appendChild(doc.createElement("button"));


  but.setAttribute("label", util0.getText("icon.panel.button.label", "Multifox (BETA)"));
  but.addEventListener("blur", function(evt) {panel.hidePopup();}, true);
  but.addEventListener("command", function(evt) {
    var w = evt.target.ownerDocument.defaultView;
    w.openUILinkIn("about:multifox", "tab", false, null, null);
  }, true);

  if (win.gBrowser.selectedBrowser.currentURI.spec === "about:multifox") {
    but.hidden = true;
  }

  win.setTimeout(function() {
    // workaround to make focus work
    but.focus();
  }, 70);

  // if (aero) {
  //panel.style.MozAppearance = "tooltip";
  //panel.style.color = "InfoText";


  panel.setAttribute("id", "multifox-popup");
  panel.addEventListener("popuphidden", closeMultifoxPopup, true);
  panel.style.padding = "0.9em 1.3em 0.9em 1.3em";
  panel.style.width = "30em";
  panel.openPopup(icon, "after_end");

  initIconPressed(icon);
}



function closeMultifoxPopup(evt) {
  var doc = evt.target.ownerDocument;

  var panel = doc.getElementById("multifox-popup");
  panel.parentNode.removeChild(panel);

  initIconNormal(getIconNode(doc));
}


function onIconHover(evt) {
  var icon = getIconNode(evt.target.ownerDocument);
  switch (evt.type) {
    case "mouseover":
      icon.style.backgroundPosition = "0pt -32px";
      icon.querySelector("label").style.textShadow = "1px 1px 1px black, 0 1px 5px white, 3px 0 5px white, 0 -1px 5px white, -3px 0 5px white";
      break;
    case "mouseout":
      icon.style.backgroundPosition = "0pt 0pt";
      icon.querySelector("label").style.textShadow = "1px 1px 1px black";
      break;
  }
}



function initIconNormal(icon) {
  icon.ownerDocument.defaultView.setTimeout(function(){
    // workaround
    // otherwise, mousedown on icon will close and open popup, even with stopPropagation
    icon.addEventListener("mousedown", openMultifoxPopup, true);
  }, 0);
  icon.addEventListener("mouseover", onIconHover, true);
  icon.addEventListener("mouseout", onIconHover, true);
  icon.style.backgroundPosition = "0pt 0pt";
  icon.querySelector("label").style.textShadow = "1px 1px 1px black";
}

function initIconPressed(icon) { // openpopup
  icon.removeEventListener("mousedown", openMultifoxPopup, true);
  icon.removeEventListener("mouseover", onIconHover, true);
  icon.removeEventListener("mouseout", onIconHover, true);
  icon.style.backgroundPosition = "0pt -64px";
  icon.querySelector("label").style.textShadow = "1px 1px 1px black, 0 1px 5px white, 3px 0 5px white, 0 -1px 5px white, -3px 0 5px white";
}

var XblContent = {

  init2: function() {
    this._enableContentXbl(true);
    xblFrameEarlyActivator.init();
  },

  shutdown2: function() {
    xblFrameEarlyActivator.shutdown();
    this._enableContentXbl(false);
  },

  _enableContentXbl: function(enable) {
    var uri = utilEx.stringToUri("chrome://multifox/content/multifox.css");
    var CiSss = Ci.nsIStyleSheetService;
    var type = CiSss.AGENT_SHEET;
    var sss = Cc["@mozilla.org/content/style-sheet-service;1"].getService(CiSss);
    if (enable) {
      if (sss.sheetRegistered(uri, type) === false) {
        sss.loadAndRegisterSheet(uri, type);
      }
    } else {
      if (sss.sheetRegistered(uri, type)) {
        sss.unregisterSheet(uri, type);
      }
    }
  },

  registerWindow2: function(win) {
    win.addEventListener("MultifoxContentEvent_ApplyXbl", this._applyXblListener, false, true);
    win.gBrowser.addTabsProgressListener(topDocumentEarlyXblActivator);
    win.gBrowser.addEventListener("pagehide", xblFrameEarlyActivator.pageHideListener, false);
    win.gBrowser.addEventListener("DOMContentLoaded", xblFrameEarlyActivator.domLoadedListener, false);//xxxdel?
  },

  unregisterWindow2: function(win) {
    win.removeEventListener("MultifoxContentEvent_ApplyXbl", this._applyXblListener, false);
    win.gBrowser.removeTabsProgressListener(topDocumentEarlyXblActivator);
    win.gBrowser.removeEventListener("pagehide", xblFrameEarlyActivator.pageHideListener, false);
    win.gBrowser.removeEventListener("DOMContentLoaded", xblFrameEarlyActivator.domLoadedListener, false);
  },

  // element bind
  _applyXblListener: function(evt) {
    evt.stopPropagation();
    var contentDoc = evt.target;
    if (contentDoc.getUserData("multifox-doc-xbl") !== null) {
      util0.log("DUP XBL " + contentDoc.location);
      return;
    }
    contentDoc.setUserData("multifox-doc-xbl", "ok", null);

    evt.preventDefault(); // flag as Multifox document
    var h = contentDoc.getElementsByTagName("head")[0];
    var w = contentDoc.documentElement.style.width;


    // remove document
    for (var idx = 0, len = m_frameNodes.length; idx < len; idx++) {
      if (m_frameNodes[idx].contentDocument === contentDoc) {
        util0.log("_applyXblListener frame removed! " + m_frameNodes[idx].src + evt.target.location.href);
        removePendingFrame(idx);
        break;
      }
    }
  }
};


const topDocumentEarlyXblActivator = {
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIWebProgressListener, Ci.nsISupportsWeakReference]),
  onStateChange: function() {},
  onSecurityChange: function() {},
  onProgressChange: function() {},
  onStatusChange: function() {},
  onLinkIconAvailable: function() {},
  onLocationChange: function(aBrowser, aWebProgress, aRequest, aLocation) {
    var contentDoc = aWebProgress.DOMWindow.document;
    if (contentDoc.getUserData("multifox-doc-xbl") !== null) {
      return;
    }

    // for frames: it fires only if DOMWindow is different from top window
    if (!contentDoc.documentElement) {
      util0.log('documentElement='+contentDoc.documentElement);
      var dummy = contentDoc.getElementsByTagName("head")[0];
      return;
    }
    var dummy = contentDoc.documentElement.style.width; // {bug} xhtml?
  }
};


// frame/iframe starts loading
//   -current page unloads (about:blank for new elements)
//   -pagehide is triggered
//     -frame node added to m_frameNodes
//     -jsd on
//   -new page starts loading
//   -jsd.topLevelHook.onCall triggered
//     -loop m_frameNodes
//       -xbl binding (triggered by jsd)
//         -new document.cookie is defined
//         -frame removed from m_frameNodes
//         -jsd off
//   -script not found by jsd
//     -xbl binding (<body>)
//       -frame removed from m_frameNodes
//       -jsd off

var m_frameNodes = null;

function removePendingFrame(idx) {
  m_frameNodes.splice(idx, 1);
  if (m_frameNodes.length === 0) {
    jsd.disable(); // disable jsd asap
  }
}

var xblFrameEarlyActivator = {

  init: function() {
    m_frameNodes = [];
    jsd.disable();
    watchXblFrame();
  },

  shutdown: function() {
    jsd.disable();
    m_frameNodes = null;
  },

  _scheduleGc: function() {
    if (this._gcTimer === null) {
      this._gcTimer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
      this._gcTimer.init(this._timerObserver, 200, Ci.nsITimer.TYPE_ONE_SHOT);
    }
  },

  _gcTimer: null,
  _timerObserver: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference]),
    observe: function(aSubject, aTopic, aData) {
      util0.log('clearPendingFrames');
      xblFrameEarlyActivator.clearPendingFrames();
    }
  },

  clearPendingFrames: function() {
    var buf = '';
    this._gcTimer = null;

    for (var idx = m_frameNodes.length - 1; idx > -1; idx--) {
      var frm = m_frameNodes[idx];
      var winFrame = frm.contentWindow;
      buf += '\nidx=' + idx;

      if (winFrame === null) {
        // win closed?
        removePendingFrame(idx);
        buf += '\nNULL ' + frm.tagName + ' src=' + frm.src;
        continue;
      }

      var browser = utilEx.findBrowser(winFrame);
      if (browser === null) {
        removePendingFrame(idx);
        buf += '\nDEL5 ' + frm.tagName + ' src=' + frm.src;
        continue;
      }

      var ifr = frm;
      do {
        var parentFrame = ifr.ownerDocument.defaultView.frameElement;
        var parentWin = parentFrame === null ? browser : parentFrame;
        var pos = parentWin.contentDocument.documentElement.compareDocumentPosition(ifr);
        if ((pos & 1) === 1) { // DOCUMENT_POSITION_DISCONNECTED=1
          removePendingFrame(idx);
          buf += '\nDEL2 '+pos + ' ' + ifr.tagName + ' src=' + ifr.src + ' ' + ifr.clientTop;
        } else {
          buf += '\nWAIT '+pos + ' ' + ifr.tagName + ' src=' + ifr.src + ' ' + ifr.clientTop;
        }
        ifr = parentFrame;
      } while (parentFrame !== null)
    }

    if (m_frameNodes.length > 0) {
      this._scheduleGc();
    }
  },

  //_clearGhostFrames: function(arr, idx) {

  touchPendingDocuments: function() {
    //util0.log('touchPendingDocuments');
    for (var idx = m_frameNodes.length - 1; idx >= 0; idx--) {
      var node = m_frameNodes[idx];
      if (node.contentDocument) {
        //var h = d.getElementsByTagName("head")[0];
        var w = node.contentDocument.documentElement.style.width;
      } else {
        removePendingFrame(idx);
        util0.log('touchPendingDocuments DEL ' + idx + ' ' + node.src);
      }
    }
  },

  domLoadedListener: {
    handleEvent: function(evt) {
      var doc = evt.target;
      var frm = doc.defaultView.frameElement;
      if (frm === null) {
        return;
      }

      var idx = m_frameNodes.indexOf(frm);
      if (idx > -1) {
        //util0.log('frame domLoadedListener/ removed ' + frm.src + '--' + doc.location.href);
        removePendingFrame(idx);
      }
    }
  },

  pageHideListener: {
    handleEvent: function(evt) {
      // this=tabbrowser
      var doc = evt.target;
      var frm = doc.defaultView.frameElement;
      if (frm === null) {
        return;
      }

      if (m_frameNodes.indexOf(frm) > -1) {
        util0.log('PageHide: frame ja em m_frameNodes '+frm.src + m_frameNodes.indexOf(frm));
        return;
      }

      util0.log('Frame pagehide add ' + frm.tagName  + ':' + frm.src + '\n' + '');
      m_frameNodes.push(frm);
      xblFrameEarlyActivator._scheduleGc();
      jsd.enable();
    }
  }
};


var jsd = {
  _hook: {
    QueryInterface: XPCOMUtils.generateQI([Ci.jsdICallHook]),
    onCall: function(frame, type) {
      if (type === Ci.jsdICallHook.TYPE_TOPLEVEL_START) {
        //util0.log('onCall');// + frame.functionName + type);
        xblFrameEarlyActivator.touchPendingDocuments();
      }
    }
  },

  _service: null,

  enable: function() {
    if (this._service === null) {
      util0.log('jsd ON');
      var k = Ci.jsdIDebuggerService;
      var srv = Cc["@mozilla.org/js/jsd/debugger-service;1"].getService(k);
      srv.topLevelHook = this._hook;
      srv.on();
      srv.flags |= k.DISABLE_OBJECT_TRACE;
      this._service = srv;
    }
  },

  disable: function() {
    util0.log('jsdoff');
    if (this._service) {
      this._service.off();
    }
    this._service = null;
  }
};


function watchXblFrame() {
  var buf='watchXblFrame';
  buf+="\n m_frameNodes.length="+ m_frameNodes.length + ' ' + (jsd._service ? jsd._service.isOn : 'jsd=null');

  var qty = m_frameNodes.length;
  for (var idx = qty - 1; idx >= 0; idx--) {
    buf+="\n" + idx + ' '+ m_frameNodes[idx].src;
  }

  util0.log(buf);

  var winEnum = utilEx.browserWindowsEnum();
  winEnum.hasMoreElements();
  var win = winEnum.getNext();
  win.setTimeout(watchXblFrame, 70700);
}
const httpObservers = {
  request: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),

    // nsIObserver
    observe: function(aSubject, aTopic, aData) {
      var httpChannel = aSubject.QueryInterface(Ci.nsIHttpChannel);
      //util0.log("Request==>" + httpChannel.URI.spec.substr(0,50));

      var domWindow = HttpContext.findWindow(httpChannel);
      if (domWindow === null) {
        return;
      }

      //chromeWindow="navigator:browser","navigator:view-source"
      //util0.log('request domWindow=' + domWindow.document.location);
      var chromeWindow = utilEx.findChromeWindow(domWindow);
      if (chromeWindow === null) {
        return;
      }

      var profileId = Profile.getIdentity(chromeWindow.document);

      if (profileId === UnknownIdentity) {
        // popup via js/window.open
        if (domWindow.opener) {
          var chromeOpener = utilEx.findChromeWindow(domWindow.opener);
          profileId = chromeOpener ? Profile.getIdentity(chromeOpener.document) : -1;
          if (profileId > UnknownIdentity) {
            Profile.save(chromeWindow, profileId);
          }
        }
        util0.log("request [" + profileId + "] id.opener="+ profileId);
      }

      if (profileId === DefaultIdentity) {
        return;
      }

      var myHeaders = HttpHeaders.fromRequest(httpChannel);
      //if (myHeaders["cookie"] !== null) {
      //  util0.log("BUG COOKIE!=NULL "+myHeaders["cookie"]);
      //}

      if (myHeaders["authorization"] !== null) {
        Components.utils.import("resource://multifox-modules/error.js");
        showError(domWindow, "HTTP Authorization");
        return;
      }

      if (profileId === UnknownIdentity) {
        httpChannel.setRequestHeader("Cookie", null, false);
        return;
      }

      var uri = toInternalUri(httpChannel.URI, profileId);
      var cook = Cookies.getCookie(false, uri);
      httpChannel.setRequestHeader("Cookie", cook, false);
    }
  },

  response: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),

    // nsIObserver
    observe: function(aSubject, aTopic, aData) {
      var httpChannel = aSubject.QueryInterface(Ci.nsIHttpChannel);
      var domWindow = HttpContext.findWindow(httpChannel);
      if (domWindow === null) {
        return;
      }

      var chromeWindow = utilEx.findChromeWindow(domWindow);
      if (chromeWindow === null) {
        return;
      }

      var profileId = Profile.getIdentity(chromeWindow.document);
      if (profileId === DefaultIdentity) {
        return;
      }


      var myHeaders = HttpHeaders.fromResponse(httpChannel);
      if (myHeaders["www-authenticate"] !== null) {
        Components.utils.import("resource://multifox-modules/error.js");
        showError(domWindow, "HTTP Authorization");
        return;
      }

      var setCookies = myHeaders["set-cookie"];
      if (setCookies === null) {
        return;
      }

      // server sent "Set-Cookie"
      httpChannel.setResponseHeader("Set-Cookie", null, false);

      if (profileId === UnknownIdentity) {
        return;
      }


      var uri = toInternalUri(httpChannel.URI, profileId);
      var multiSetCookies = new SetCookieParser(setCookies, true);

      multiSetCookies.forEach(function(c) {
        var realDomain = c.meta.domain;
        if (realDomain.length > 0) {
          c.meta.domain = cookieInternalDomain(realDomain, profileId);
        }
      });

      var internalCookies = multiSetCookies.toHeader();
      Cookies.setCookie(false, uri, internalCookies);
    }
  }
};


var HttpContext = {
  findWindow: function(request) {
    var context = this._loadContext(request);
    if (context) {
      if (!context.isContent) {
        util0.log('CONTEXT=CHROME => ' + request.URI.spec);
      }
      return context.isContent ? context.associatedWindow : null;
    }

    // nsILoadContext doesn't work with CSS (background image)
    var progress = this._webProgress(request);
    if (progress) {
      //Multifox@06:37:15>BUG gbrowser=null / win still loading? chrome://browser/content/browser.xul[object ChromeWindow]
      var isChrome = progress.DOMWindow instanceof Components.interfaces.nsIDOMChromeWindow;
      //util0.log('progress ' + isChrome + progress.DOMWindow.document.location);//progress.DOMWindow.document.documentURIObject +
      return isChrome ? null : progress.DOMWindow;
    }

    util0.log('CONTEXT FAIL => ' + request.URI.spec);
    return null;
  },

  _loadContext: function(request) {
    try {
      return request
              .queryInterface(Ci.nsIChannel)
              .notificationCallbacks
              .getInterface(Ci.nsILoadContext);
    } catch (ex) {}

    try {
      return request
              .loadGroup
              .notificationCallbacks
              .getInterface(Ci.nsILoadContext);
    } catch (ex) {}

    return null;
  },

  _webProgress: function(request) {
    if (request.notificationCallbacks) {
      try {
        return request
                .notificationCallbacks
                .getInterface(Ci.nsIWebProgress);
      } catch (ex) {}
    }

    try {
      if (request.loadGroup && request.loadGroup.groupObserver) {
        return request
                .loadGroup
                .groupObserver
                .QueryInterface(Ci.nsIWebProgress);
      }
    } catch (ex) {}

    return null;
  }
};


const HttpHeaders = {
  visitLoop: {
    values: null,
    visitHeader: function(name, value) {
      var n = name.toLowerCase();
      if (n in this.values) {
        this.values[n] = value;
      }
    }
  },

  fromRequest: function(request) {
    var nameValues = {
      "cookie": null, //for debug only
      "authorization": null
    }
    this.visitLoop.values = nameValues;
    request.visitRequestHeaders(this.visitLoop);
    return nameValues;
  },

  fromResponse: function(response) {
    var nameValues = {
      "set-cookie": null,
      "www-authenticate": null
    }
    this.visitLoop.values = nameValues;
    response.visitResponseHeaders(this.visitLoop);
    return nameValues;
  }
};
var m_documentCookie = null;
var Cookies = {
  _service: null,
  _prefs: null,

  init3: function() {
    m_documentCookie = {};
    this._service = Cc["@mozilla.org/cookieService;1"].getService().QueryInterface(Ci.nsICookieService);
    this._prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch2);
    this.obs.behavior = this._prefs.getIntPref("network.cookie.cookieBehavior");
    this._prefs.addObserver("network.cookie.cookieBehavior", this.obs, false);
  },


  obs: {
    behavior: -1,

    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
    // aTopic=nsPref:changed aData=network.cookie.cookieBehavior
    observe: function(aSubject, aTopic, aData) {
      this.behavior = aSubject
                      .QueryInterface(Components.interfaces.nsIPrefBranch)
                      .getIntPref("network.cookie.cookieBehavior");
      util0.log('pref! ' + aSubject + aTopic + aData + this.behavior);
    }

  },

  shutdown3: function() {
    m_documentCookie = null;
    this._service = null;
    this._prefs.removeObserver("network.cookie.cookieBehavior", this.obs);
    this._prefs = null;
  },

  regWindow: function(sessionId) {
    var cook;
    if (("id" + sessionId) in m_documentCookie) {
      cook = m_documentCookie["id" + sessionId];
    } else {
      cook = new documentCookieListeners(sessionId);
      m_documentCookie["id" + sessionId] = cook;
    }
    return cook;
  },

  removeWindow: function(idw, sessions) {
    var cook = m_documentCookie["id" + idw];
    if (sessions.indexOf(idw) === -1) {
      delete m_documentCookie["id" + idw];
    }
    return cook;
  },

  setCookie: function(fromJs, uri, val) {
    if (this.obs.behavior === 0) {
      this._setCookie(fromJs, uri, val);
      return;
    }

    var p = this._prefs;
    var name = "network.cookie.cookieBehavior";
    p.removeObserver(name, this.obs);
    p.setIntPref(name, 0);
    this._setCookie(fromJs, uri, val);
    p.setIntPref(name, this.obs.behavior);
    p.addObserver(name, this.obs, false);
  },

  getCookie: function(fromJs, uri) {
    if (this.obs.behavior === 0) {
      return this._getCookie(fromJs, uri);
    }

    var p = this._prefs;
    var name = "network.cookie.cookieBehavior";
    p.removeObserver(name, this.obs);
    p.setIntPref(name, 0);
    var cookie = this._getCookie(fromJs, uri);
    p.setIntPref(name, this.obs.behavior);
    p.addObserver(name, this.obs, false);
    return cookie;
  },

  _setCookie: function(fromJs, uri, val) {
    if (fromJs) {
      this._service.setCookieString(uri,
                                    null,
                                    val,
                                    null);
    } else {
      //setCookieString doesn't work for httponly cookies
      this._service.setCookieStringFromHttp(uri,   // aURI
                                            null,  // aFirstURI
                                            null,  // aPrompt
                                            val,   // aCookie
                                            null,  // aServerTime
                                            null); // aChannel
    }
  },

  _getCookie: function(fromJs, uri) {
    if (fromJs) {
      return this._service.getCookieString(uri,
                                           null);
    } else {
      return this._service.getCookieStringFromHttp(uri,   // aURI
                                                   null,  // aFirstURI
                                                   null); // aChannel
    }
  }
};



function documentCookieListeners(id) {
  this.cookieGetter = new documentCookieGetter(id);
  this.cookieSetter = new documentCookieSetter(id);
}

function documentCookieGetter(id) {
  this.id = id;
}

documentCookieGetter.prototype = {
  id: -1,
  handleEvent: function(evt) {
    evt.stopPropagation();
    //utilEx.dbg('cookie getter', this.id, evt.target.location.href);

    var profileId = this.id;
    switch (profileId) {
      case UnknownIdentity:
        break;
      case DefaultIdentity:
        utilEx.thrw.myThrow("documentCookieSetter " + profileId);
        break;
      default:
        var contentDoc = evt.target;
        var contentUrl = contentDoc.location.href;
        var uri = toInternalUri(utilEx.stringToUri(contentUrl), this.id);

        var cookie2 = Cookies.getCookie(true, uri);
        var cookie = cookie2 === null ? "" : cookie2;

        var myEvent = contentDoc.createEvent("MessageEvent");
        myEvent.initMessageEvent(   // send cookie value to content
          "MultifoxContentEvent_CookieGetter_Reply",
          true,
          false,
          cookie,
          contentUrl,
          "",
          contentDoc.defaultView);
        var success = contentDoc.dispatchEvent(myEvent);
        return;
    }
  }
};

function documentCookieSetter(id) {
  this.id = id;
}

documentCookieSetter.prototype = {
  id: -1,
  handleEvent: function(evt) {
    evt.stopPropagation();
    var profileId = this.id;
    //utilEx.dbg('cookie setter', this.id, evt.target.location.href, evt.data);

    switch (profileId) {
      case UnknownIdentity:
        break;
      case DefaultIdentity:
        utilEx.thrw.myThrow("documentCookieSetter " + profileId);
        break;
      default:
        var contentDoc = evt.target;
        var contentUrl = contentDoc.location.href;
        var unsafeCookie = evt.data;
        var uri = toInternalUri(utilEx.stringToUri(contentUrl), profileId);
        Cookies.setCookie(true, uri, unsafeCookie);
        break;
    }
  }
};



function toInternalUri(uri, sessionId) {
  var u = uri.clone();
  if (sessionId > DefaultIdentity) {
    u.host = cookieInternalDomain(u.host, sessionId);
  } else {
    utilEx.thrw.myThrow("invalid profile " + sessionId);
  }
  return u;
}


function cookieInternalDomain(domain, id) {
  return domain + ".multifox-profile-" + id;
}


function splitValueName(cookie) {
  var idx = cookie.indexOf("=");
  if (idx === -1) {
    return [cookie.trim(), null];
  }

  // 'a=bcd=e'.split('=',2) returns [a,bcd]
  //   'abcde'.split('=',2) returns [abcde]


  // MY =
  // 012^-----idx=3 length=4

  // MY =a:1=6
  // 012^-----idx=3 length=9

  var pair = ["", ""];
  pair[0] = cookie.substring(0, idx).trim();//trim(cookie.substring(0, idx));
  idx++;
  if (idx < cookie.length) {
    pair[1] = cookie.substring(idx);
  }

  return pair;
}



function toLines(txt) {
  return txt.split(/\r\n|\r|\n/);
}


function SetCookieParser(cookieHeader, hasMeta) {
  this.m_hasMeta = hasMeta;
  this.m_singleCookies = [];
  if (cookieHeader !== null) {
    var lines = toLines(cookieHeader);
    var len = lines.length;
    if (hasMeta) {
      for (var idx = 0; idx < len; idx++) {
        this.parseCookieLineMeta(lines[idx]);
      }
    } else {
      for (var idx = 0; idx < len; idx++) {
        this.parseCookieLine(lines[idx]);
      }
    }
  }
}



SetCookieParser.prototype = {
  parseCookieLineMeta: function(headerLine) {
    const values = {};//xxxbug set-cookie ==> one name=value
    const meta = new CookieMeta();
    const items = headerLine.split(";");

    for (var idx = 0, len = items.length; idx < len; idx++) {
      var pair = splitValueName(items[idx]);
      var name = pair[0];
      var value = pair[1]; // null ==> name=HttpOnly, secure etc

      if (idx === 0) {
        if (name.length > 0 && value !== null) {
          values[name] = value;
        } else {
          utilEx.dbg('m_singleCookies invalid', name, value, headerLine);
          break;
        }
        continue;
      }


      switch (name.toLowerCase()) {
        /*
        case "expires":
          meta.expires = value;
          continue;
        case "domain":
          meta.domain = value;
          continue;
        case "path":
          meta.path = value;
          continue;
        */
        case "expires":
        case "domain":
        case "path":
          if (value !== null) {
            meta[name.toLowerCase()] = value;
          }
          continue;

        case "secure":
          if (value === null) {
            meta.secureOnly = true;
          //} else {
          //  values[name] = value;
          }
          break;
        case "httponly":
          if (value === null) {
            meta.httpOnly = true;
          //} else {
          //  values[name] = value;
          }
          break;
        //default:
        //  values[name] = value;
        //  break;
      }
    }


    for (var name in values) {
      var my = new SimpleCookie(name, values[name], meta);
      this.m_singleCookies.push(my);
    }
  },

  parseCookieLine: function(headerLine) {
    var items = headerLine.split(';');
    for (var idx = 0, len = items.length; idx < len; idx++) {
      var pair = splitValueName(items[idx]);
      var my = new SimpleCookie(pair[0], pair[1], null);
      this.m_singleCookies.push(my);
    }
  },

  toHeader: function() {
    function toHeaderCore() {
      var len = c.length;
      var buf = []; //new Array(len);
      for (var idx = 0, len = c.length; idx < len; idx++) {
        if (c[idx].value !== null) {
          buf[idx] = c[idx].toHeaderLine();
        }
      }
      return buf;
    }
    var c = this.m_singleCookies;
    return this.m_hasMeta
      ? toHeaderCore().join('\n')
      : toHeaderCore().join(';');
  },

  getCookie: function(name) {
    var c = this.m_singleCookies;
    for (var idx = 0, len = c.length; idx < len; idx++) {
      if (name === c[idx].name()) {
        return c[idx];
      }
    }
    return null;
  },

  forEach: function(fn) {
    var c = this.m_singleCookies;
    for (var idx = 0, len = c.length; idx < len; idx++) {
      fn(c[idx]);
    }
  }
};

function CookieMeta() {}

CookieMeta.prototype = {
  path:       "",
  domain:     "",
  expires:    "",
  httpOnly:   false,
  secureOnly: false,
  clone: function() {
    var c = new CookieMeta();
    c.domain = this.domain;
    c.path = this.path;
    c.expires = this.expires;
    c.httpOnly = this.httpOnly;
    c.secureOnly = this.secureOnly;
    return c;
  }
};

function SimpleCookie(m_name, m_value, m_meta) { // SingleCookie
  this.name = function() {
    return m_name;
  }
  this.value = m_value;
  this.meta = m_meta;
}

SimpleCookie.prototype = {
  clone: function() {
    return new SimpleCookie(this.name(), this.value, this.meta.clone());
  },
  toHeaderLine: function() {
    var m = this.meta;
    if (m === null) {
      return this.name() + "=" + this.value;
    }

    var buf = [this.name() + "=" + this.value];

    if (m.expires.length > 0) {
      buf.push("EXPIRES=" + m.expires);
    }

    if (m.domain.length > 0) {
      buf.push("DOMAIN=" + m.domain);
    }

    if (m.path.length > 0) {
      buf.push("PATH=" + m.path);
    }

    if (m.secureOnly) {
      buf.push("SECURE");
    }

    if (m.httpOnly) {
      buf.push("HTTPONLY");
    }

    return buf.join("; ");
  }
};
const utilEx = {

  dbg: function() {
    var buf = "Debug @";// + this.dbg.caller + "\n";// + utilEx.thrw._toString(false, Components.stack) + "===";
    for (var idx = 0, len = arguments.length; idx < len; idx++) {
      buf += "\n" + arguments[idx];// + "=" + this.caller[arguments[idx]];
    }

    buf += "\n====\n" + utilEx.thrw._toString(false, Components.stack);
    util0.log(buf);
  },

  findChromeWindow: function(domWindow) {
    var docShellTreeItem = domWindow
                            .QueryInterface(Ci.nsIInterfaceRequestor)
                            .getInterface(Ci.nsIWebNavigation)
                            .QueryInterface(Ci.nsIDocShellTreeItem);

    var win = docShellTreeItem
              .rootTreeItem
              .QueryInterface(Ci.nsIInterfaceRequestor)
              .getInterface(Ci.nsIDOMWindow)
              .wrappedJSObject; //allow access to gBrowser etc

    if (win instanceof Ci.nsIDOMChromeWindow) {
      return win;
    }

    // it doesn't work for some popups (docShellTreeItem.treeOwner is null)
    // try a [probably] slower method
    //utilEx.dbg('win NOT nsIDOMChromeWindow', domWindow.document.location, win.document.location);
    var browser = this.findBrowser2(domWindow);
    return browser ? browser.ownerDocument.defaultView : null;
  },

  browserWindowsEnum: function() {
    return Cc["@mozilla.org/appshell/window-mediator;1"]
            .getService(Ci.nsIWindowMediator)
            .getEnumerator("navigator:browser");
  },

  findBrowser2: function(domWindow) {
    var doc = domWindow.top.document;
    var winEnum = utilEx.browserWindowsEnum();
    while (winEnum.hasMoreElements()) {
      var chrWin = winEnum.getNext();
      var idx = chrWin.gBrowser.getBrowserIndexForDocument(doc);
      if (idx > -1) {
        return chrWin.gBrowser.browsers[idx];
      }
    }
    return null;
  },

  findBrowser: function(domWindow) {
    var win = this.findChromeWindow(domWindow);
    if (win) {
      var idx = win.gBrowser.getBrowserIndexForDocument(domWindow.top.document);
      if (idx > -1) {
        return win.gBrowser.browsers[idx];
      }
    }
    utilEx.dbg('findBrowser=null', domWindow.document.location, win, this.findBrowser2(domWindow));
    return null;
  },

/*

  _findViewSourceWindow: function(domWindow) {
    var doc = domWindow.top.document;
    var winEnum = utilEx.viewSourceWindowsEnum();
    while (winEnum.hasMoreElements()) {
      var chrWin = winEnum.getNext();
      var browser = chrWin.document.getElementById("content");
      if (browser.contentDocument === doc) {
        return chrWin;
      }
    }
    return null;
  },

  viewSourceWindowsEnum: function() {
    return Cc["@mozilla.org/appshell/window-mediator;1"]
            .getService(Ci.nsIWindowMediator)
            .getEnumerator("navigator:view-source");
  },
*/
  appName: function() {
    return Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULAppInfo).name;
  },

  stringToUri: function(spec) {
    try {
      return Cc["@mozilla.org/network/io-service;1"]
              .getService(Ci.nsIIOService)
              .newURI(spec, null, null);
    } catch (ex) {
      return null;
    }
  },


  thrw: {

    myThrow: function(txt) {
      util0.log('Stack: ' + txt + '\n' + this._toString(false, Components.stack));
      throw txt;
    },

    _toArray: function(webOnly, theStack) {
      var allItems = [];
      if (!theStack) {
        //$.moz.prompt('theStack undef\n'+$.stack.toString(false, Components.stack))
        return allItems;
      }

      function StackItem() {}
      StackItem.prototype = {
        lang: '?',
        filename: '?',
        lineNumber: -1,
        name: '?'
      };

      // aparentemente  um bug que ocorre as vezes, todas as propriedades sao undefined
      if (theStack.languageName == undefined) {
        //$.moz.prompt('myStack\n'+myStack +  '\n\ntheStack\n'+theStack + '\n\nallItems\n' + allItems.join('\n') + '\n\n\n'+$.stack.toString(false, Components.stack))
        var item = new StackItem();
        item.name = theStack.toString();
        allItems.push(item);
      }

      //for (var myStack = theStack; myStack !== null; myStack = myStack.caller) {
      //    $.moz.prompt('myStack\n'+myStack);
      //}
      //var b='';for (var myStack = Components.stack; myStack; myStack = myStack.caller) {b+='\n'+myStack.filename;}alert(target.opener+b)
      // myStack.caller  quase sempre null, mas algumas vezes  undefined...
      for (var myStack = theStack; myStack; myStack = myStack.caller) {
        if (webOnly) {
          var n = myStack.filename;
          if (n === null) {
            continue;
          }

          /*
          var sch = Cc["@mozilla.org/network/io-service;1"]
                        .getService(Ci.nsIIOService)
                        .extractScheme(n);
          */

          if ((n.indexOf('http:') !== 0) && (n.indexOf('https:') !== 0)) {
            continue;
          }
        }


        var item = new StackItem();
        allItems.push(item);
        item.lang = myStack.languageName;
        item.filename = myStack.filename;
        item.lineNumber = myStack.lineNumber; // Valid line numbers begin at '1'. '0' indicates unknown.
        item.name = myStack.name;
      }

      allItems.reverse();
      return allItems;
    },

    _toString: function(webOnly, theStack) {
      var arr = this._toArray(webOnly, theStack);
      var lines = new Array(arr.length);
      for (var idx = arr.length - 1, idx2 = 0; idx > -1; idx--) {
        var lang = webOnly ? '' : arr[idx].lang + ' ';
        lines[idx2] = '[' + (idx + 1) + '] '
                    + lang
                    + arr[idx].filename + ' '
                    + arr[idx].name + ' ('
                    + arr[idx].lineNumber + ')';
        idx2++;
      }
      return lines.join('\n');
    }

  }
};
function getAboutUriChannel(uri) {
  var channel = Cc["@mozilla.org/network/io-service;1"]
                  .getService(Ci.nsIIOService)
                  .newChannel("chrome://multifox/content/about.html", null, null);
  channel.originalURI = uri;
  return channel;
}
